# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) Vispy Development Team. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
# -----------------------------------------------------------------------------

# Adapted from Pyglet

import atexit
from functools import partial
import struct

from ctypes import (windll, Structure, POINTER, byref, WINFUNCTYPE,
                    c_uint, c_float, c_int, c_ulong, c_uint64,
                    c_void_p, c_uint32, c_wchar, c_wchar_p)
from ctypes.wintypes import (LONG, BYTE, HFONT, HGDIOBJ, BOOL, UINT, INT,
                             DWORD, LPARAM)

try:
    import _winreg as winreg
except ImportError:
    import winreg  # noqa, analysis:ignore

_64_bit = (8 * struct.calcsize("P")) == 64

LF_FACESIZE = 32
FW_BOLD = 700
FW_NORMAL = 400
ANTIALIASED_QUALITY = 4
FontStyleBold = 1
FontStyleItalic = 2
UnitPixel = 2
UnitPoint = 3
DEFAULT_CHARSET = 1
ANSI_CHARSET = 0
TRUETYPE_FONTTYPE = 4
GM_ADVANCED = 2
CSIDL_FONTS = 0x0014

PixelFormat24bppRGB = 137224
PixelFormat32bppRGB = 139273
PixelFormat32bppARGB = 2498570

DriverStringOptionsCmapLookup = 1
DriverStringOptionsRealizedAdvance = 4
TextRenderingHintAntiAlias = 4
TextRenderingHintAntiAliasGridFit = 3
ImageLockModeRead = 1
StringFormatFlagsMeasureTrailingSpaces = 0x00000800
StringFormatFlagsNoClip = 0x00004000
StringFormatFlagsNoFitBlackBox = 0x00000004

INT_PTR = c_int
REAL = c_float
TCHAR = c_wchar
UINT32 = c_uint32
HDC = c_void_p
PSTR = c_uint64 if _64_bit else c_uint

HORZSIZE = 4
VERTSIZE = 6

HORZRES = 8
VERTRES = 10


# gdi32

class POINT(Structure):
    _fields_ = [('x', LONG), ('y', LONG)]


class RECT(Structure):
    _fields_ = [('left', LONG), ('top', LONG),
                ('right', LONG), ('bottom', LONG)]


class PANOSE(Structure):
    _fields_ = [
        ('bFamilyType', BYTE), ('bSerifStyle', BYTE), ('bWeight', BYTE),
        ('bProportion', BYTE), ('bContrast', BYTE), ('bStrokeVariation', BYTE),
        ('bArmStyle', BYTE), ('bLetterform', BYTE), ('bMidline', BYTE),
        ('bXHeight', BYTE)]


class TEXTMETRIC(Structure):
    _fields_ = [
        ('tmHeight', LONG), ('tmAscent', LONG), ('tmDescent', LONG),
        ('tmInternalLeading', LONG), ('tmExternalLeading', LONG),
        ('tmAveCharWidth', LONG), ('tmMaxCharWidth', LONG),
        ('tmWeight', LONG), ('tmOverhang', LONG),
        ('tmDigitizedAspectX', LONG), ('tmDigitizedAspectY', LONG),
        ('tmFirstChar', TCHAR), ('tmLastChar', TCHAR),
        ('tmDefaultChar', TCHAR), ('tmBreakChar', TCHAR),
        ('tmItalic', BYTE), ('tmUnderlined', BYTE),
        ('tmStruckOut', BYTE), ('tmPitchAndFamily', BYTE),
        ('tmCharSet', BYTE)]


class OUTLINETEXTMETRIC(Structure):
    _fields_ = [
        ('otmSize', UINT), ('otmTextMetrics', TEXTMETRIC),
        ('otmMysteryBytes', BYTE), ('otmPanoseNumber', PANOSE),
        ('otmMysteryByte', BYTE),
        ('otmfsSelection', UINT), ('otmfsType', UINT),
        ('otmsCharSlopeRise', INT), ('otmsCharSlopeRun', INT),
        ('otmItalicAngle', INT), ('otmEMSquare', UINT), ('otmAscent', INT),
        ('otmDescent', INT), ('otmLineGap', UINT), ('otmsCapEmHeight', UINT),
        ('otmsXHeight', UINT), ('otmrcFontBox', RECT), ('otmMacAscent', INT),
        ('otmMacDescent', INT), ('otmMacLineGap', UINT),
        ('otmusMinimumPPEM', UINT), ('otmptSubscriptSize', POINT),
        ('otmptSubscriptOffset', POINT), ('otmptSuperscriptSize', POINT),
        ('otmptSuperscriptOffset', POINT), ('otmsStrikeoutSize', UINT),
        ('otmsStrikeoutPosition', INT), ('otmsUnderscoreSize', INT),
        ('otmsUnderscorePosition', INT), ('otmpFamilyName', PSTR),
        ('otmpFaceName', PSTR), ('otmpStyleName', PSTR),
        ('otmpFullName', PSTR), ('junk', (BYTE) * 1024)]  # room for strs


class LOGFONT(Structure):
    _fields_ = [
        ('lfHeight', LONG), ('lfWidth', LONG), ('lfEscapement', LONG),
        ('lfOrientation', LONG), ('lfWeight', LONG), ('lfItalic', BYTE),
        ('lfUnderline', BYTE), ('lfStrikeOut', BYTE), ('lfCharSet', BYTE),
        ('lfOutPrecision', BYTE), ('lfClipPrecision', BYTE),
        ('lfQuality', BYTE), ('lfPitchAndFamily', BYTE),
        ('lfFaceName', (TCHAR * LF_FACESIZE))]


gdi32 = windll.gdi32

gdi32.CreateFontIndirectW.restype = HFONT
gdi32.CreateFontIndirectW.argtypes = [POINTER(LOGFONT)]

gdi32.SelectObject.restype = HGDIOBJ
gdi32.SelectObject.argtypes = [HDC, HGDIOBJ]

gdi32.SetGraphicsMode.restype = INT
gdi32.SetGraphicsMode.argtypes = [HDC, INT]

gdi32.GetTextMetricsW.restype = BOOL
gdi32.GetTextMetricsW.argtypes = [HDC, POINTER(TEXTMETRIC)]

FONTENUMPROC = WINFUNCTYPE(INT, POINTER(LOGFONT), POINTER(TEXTMETRIC),
                           DWORD, c_void_p)
gdi32.EnumFontFamiliesExW.restype = INT
gdi32.EnumFontFamiliesExW.argtypes = [HDC, POINTER(LOGFONT),
                                      FONTENUMPROC, LPARAM, DWORD]

gdi32.GetOutlineTextMetricsW.restype = UINT
gdi32.GetOutlineTextMetricsW.argtypes = [HDC, UINT,
                                         POINTER(OUTLINETEXTMETRIC)]


gdi32.GetDeviceCaps.argtypes = [HDC, INT]
gdi32.GetDeviceCaps.restype = INT

user32 = windll.user32

user32.GetDC.restype = HDC  # HDC
user32.GetDC.argtypes = [UINT32]  # HWND

user32.ReleaseDC.argtypes = [c_void_p, HDC]

try:
    user32.SetProcessDPIAware.argtypes = []
except AttributeError:
    pass  # not present on XP


# gdiplus

class GdiplusStartupInput(Structure):
    _fields_ = [
        ('GdiplusVersion', UINT32), ('DebugEventCallback', c_void_p),
        ('SuppressBackgroundThread', BOOL), ('SuppressExternalCodecs', BOOL)]


class GdiplusStartupOutput(Structure):
    _fields = [('NotificationHookProc', c_void_p),
               ('NotificationUnhookProc', c_void_p)]

gdiplus = windll.gdiplus

gdiplus.GdipCreateFontFamilyFromName.restype = c_int
gdiplus.GdipCreateFontFamilyFromName.argtypes = [c_wchar_p, c_void_p, c_void_p]

gdiplus.GdipNewPrivateFontCollection.restype = c_int
gdiplus.GdipNewPrivateFontCollection.argtypes = [c_void_p]

gdiplus.GdipPrivateAddFontFile.restype = c_int
gdiplus.GdipPrivateAddFontFile.argtypes = [c_void_p, c_wchar_p]

gdiplus.GdipGetFamilyName.restype = c_int
gdiplus.GdipGetFamilyName.argtypes = [c_void_p, c_wchar_p, c_int]


def gdiplus_init():
    token = c_ulong()
    startup_in = GdiplusStartupInput()
    startup_in.GdiplusVersion = 1
    startup_out = GdiplusStartupOutput()
    gdiplus.GdiplusStartup(byref(token), byref(startup_in), byref(startup_out))
    atexit.register(partial(gdiplus.GdiplusShutdown, token))

gdiplus_init()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2014, Nicolas P. Rougier. All rights reserved.
# Distributed under the terms of the new BSD License.
# -----------------------------------------------------------------------------
import unittest
import numpy as np

from vispy.testing import run_tests_if_main
from vispy.gloo.buffer import (Buffer, DataBuffer, DataBufferView, 
                               VertexBuffer, IndexBuffer)


# -----------------------------------------------------------------------------
class BufferTest(unittest.TestCase):

    # Default init
    # ------------
    def test_init_default(self):
        """ Test buffer init"""
        
        # No data
        B = Buffer()
        assert B.nbytes == 0
        glir_cmd = B._glir.clear()[-1]
        assert glir_cmd[0] == 'CREATE'
        
        # With data
        data = np.zeros(100)
        B = Buffer(data=data)
        assert B.nbytes == data.nbytes
        glir_cmd = B._glir.clear()[-1]
        assert glir_cmd[0] == 'DATA'
        
        # With nbytes
        B = Buffer(nbytes=100)
        assert B.nbytes == 100
        glir_cmd = B._glir.clear()[-1]
        assert glir_cmd[0] == 'SIZE'
        
        # Wrong data
        self.assertRaises(ValueError, Buffer, data, 4)
        self.assertRaises(ValueError, Buffer, data, data.nbytes)

    # Check setting the whole buffer clear pending operations
    # -------------------------------------------------------
    def test_set_whole_data(self):
        data = np.zeros(100)
        B = Buffer(data=data)
        B._glir.clear()
        B.set_data(data=data)
        glir_cmds = B._glir.clear()
        assert len(glir_cmds) == 2
        assert glir_cmds[0][0] == 'SIZE'
        assert glir_cmds[1][0] == 'DATA'
    
        # And sub data
        B.set_subdata(data[:50], 20)
        glir_cmds = B._glir.clear()
        assert len(glir_cmds) == 1
        assert glir_cmds[0][0] == 'DATA'
        assert glir_cmds[0][2] == 20  # offset
        
        # And sub data
        B.set_subdata(data)
        glir_cmds = B._glir.clear()
        assert glir_cmds[-1][0] == 'DATA'
        
        # Wrong ways to set subdata
        self.assertRaises(ValueError, B.set_subdata, data[:50], -1)  # neg
        self.assertRaises(ValueError, B.set_subdata, data, 10)  # no fit
        
    # Check stored data is data
    # -------------------------
    def test_data_storage(self):
        data = np.zeros(100)
        B = Buffer(data=data)
        B.set_data(data=data[:50], copy=False)
        glir_cmd = B._glir.clear()[-1]
        assert glir_cmd[-1].base is data

    # Check setting oversized data
    # ----------------------------
    def test_oversized_data(self):
        data = np.zeros(10)
        B = Buffer(data=data)
        # with self.assertRaises(ValueError):
        #    B.set_data(np.ones(20))
        self.assertRaises(ValueError, B.set_subdata, np.ones(20), offset=0)

    # Check negative offset
    # ---------------------
    def test_negative_offset(self):
        data = np.zeros(10)
        B = Buffer(data=data)
        # with self.assertRaises(ValueError):
        #    B.set_data(np.ones(1), offset=-1)
        self.assertRaises(ValueError, B.set_subdata, np.ones(1), offset=-1)

    # Check offlimit offset
    # ---------------------
    def test_offlimit_offset(self):
        data = np.zeros(10)
        B = Buffer(data=data)
        # with self.assertRaises(ValueError):
        #    B.set_data(np.ones(1), offset=10 * data.dtype.itemsize)
        self.assertRaises(ValueError, B.set_subdata,
                          np.ones(1), offset=10 * data.dtype.itemsize)

    # Buffer size
    # -----------
    def test_buffer_size(self):
        data = np.zeros(10)
        B = Buffer(data=data)
        assert B.nbytes == data.nbytes

    # Resize
    # ------
    def test_buffer_resize(self):
        data = np.zeros(10)
        B = Buffer(data=data)
        data = np.zeros(20)
        B.set_data(data)
        assert B.nbytes == data.nbytes


# -----------------------------------------------------------------------------
class DataBufferTest(unittest.TestCase):

    # Default init
    # ------------
    def test_default_init(self):
        # Check default storage and copy flags
        data = np.ones(100)
        B = DataBuffer(data)
        assert B.nbytes == data.nbytes
        assert B.offset == 0
        assert B.size == 100
        assert B.itemsize == data.itemsize
        assert B.stride == data.itemsize
        assert B.dtype == data.dtype
        
        # Given data must be actual numeric data
        self.assertRaises(TypeError, DataBuffer, 'this is not nice data')
    
    # Default init with structured data
    # ---------------------------------
    def test_structured_init(self):
        # Check structured type
        dtype = np.dtype([('position', np.float32, 3),
                          ('texcoord', np.float32, 2),
                          ('color',    np.float32, 4)])
        data = np.zeros(10, dtype=dtype)
        B = DataBuffer(data)
        assert B.nbytes == data.nbytes
        assert B.offset == 0
        assert B.size == 10
        assert B.itemsize == data.itemsize
        assert B.stride == data.itemsize
        assert B.dtype == data.dtype
        
    # No CPU storage
    # --------------
    def test_no_storage_copy(self):
        data = np.ones(100, np.float32)
        B = DataBuffer(data)
        assert B.stride == 4

    # Wrong storage
    # -------------
    def test_non_contiguous_storage(self):
        # Ask to have CPU storage and to use data as storage
        # Not possible since data[::2] is not contiguous
        data = np.ones(100, np.float32)
        data_given = data[::2]
        
        B = DataBuffer(data_given)
        assert B.stride == 4*2
    
    # Get buffer field
    # ----------------
    def test_getitem_field(self):
        dtype = np.dtype([('position', np.float32, 3),
                          ('texcoord', np.float32, 2),
                          ('color',    np.float32, 4)])
        data = np.zeros(10, dtype=dtype)
        B = DataBuffer(data)

        Z = B["position"]
        assert Z.nbytes == 10 * 3 * np.dtype(np.float32).itemsize
        assert Z.offset == 0
        assert Z.size == 10
        assert Z.itemsize == 3 * np.dtype(np.float32).itemsize
        assert Z.stride == (3 + 2 + 4) * np.dtype(np.float32).itemsize
        assert Z.dtype == (np.float32, 3)

        Z = B["texcoord"]
        assert Z.nbytes == 10 * 2 * np.dtype(np.float32).itemsize
        assert Z.offset == 3 * np.dtype(np.float32).itemsize
        assert Z.size == 10
        assert Z.itemsize == 2 * np.dtype(np.float32).itemsize
        assert Z.stride == (3 + 2 + 4) * np.dtype(np.float32).itemsize
        assert Z.dtype == (np.float32, 2)

        Z = B["color"]
        assert Z.nbytes == 10 * 4 * np.dtype(np.float32).itemsize
        assert Z.offset == (2 + 3) * np.dtype(np.float32).itemsize
        assert Z.size == 10
        assert Z.itemsize == 4 * np.dtype(np.float32).itemsize
        assert Z.stride == (3 + 2 + 4) * np.dtype(np.float32).itemsize
        assert Z.dtype == (np.float32, 4)

    # Get view via index
    # ------------------
    def test_getitem_index(self):
        dtype = np.dtype([('position', np.float32, 3),
                          ('texcoord', np.float32, 2),
                          ('color',    np.float32, 4)])
        data = np.zeros(10, dtype=dtype)
        B = DataBuffer(data)
        Z = B[0:1]
        assert Z.base == B
        assert Z.id == B.id
        assert Z.nbytes == 1 * (3 + 2 + 4) * np.dtype(np.float32).itemsize
        assert Z.offset == 0
        assert Z.size == 1
        assert Z.itemsize == (3 + 2 + 4) * np.dtype(np.float32).itemsize
        assert Z.stride == (3 + 2 + 4) * np.dtype(np.float32).itemsize
        assert Z.dtype == B.dtype
        assert 'DataBufferView' in repr(Z)
        
        # There's a few things we cannot do with a view
        self.assertRaises(RuntimeError, Z.set_data, data)
        self.assertRaises(RuntimeError, Z.set_subdata, data)
        self.assertRaises(RuntimeError, Z.resize_bytes, 20)
        self.assertRaises(RuntimeError, Z.__getitem__, 3)
        self.assertRaises(RuntimeError, Z.__setitem__, 3, data)
    
    # View get invalidated when base is resized
    # -----------------------------------------
    def test_invalid_view_after_resize(self):
        dtype = np.dtype([('position', np.float32, 3),
                          ('texcoord', np.float32, 2),
                          ('color',    np.float32, 4)])
        data = np.zeros(10, dtype=dtype)
        B = DataBuffer(data)
        Y = B['position']
        Z = B[5:]
        B.resize_bytes(5)
        assert Y._valid is False
        assert Z._valid is False

    # View get invalidated after setting oversized data
    # -------------------------------------------------
    def test_invalid_view_after_set_data(self):
        dtype = np.dtype([('position', np.float32, 3),
                          ('texcoord', np.float32, 2),
                          ('color',    np.float32, 4)])
        data = np.zeros(10, dtype=dtype)
        B = DataBuffer(data)
        Z = B[5:]
        B.set_data(np.zeros(15, dtype=dtype))
        assert Z._valid is False

    # Set data on base buffer : ok
    # ----------------------------
    def test_set_data_base(self):
        dtype = np.dtype([('position', np.float32, 3),
                          ('texcoord', np.float32, 2),
                          ('color',    np.float32, 4)])
        data = np.zeros(10, dtype=dtype)
        B = DataBuffer(data)
        B.set_data(data)
        last_cmd = B._glir.clear()[-1]
        assert last_cmd[0] == 'DATA'
        
        # Extra kwargs are caught
        self.assertRaises(TypeError, B.set_data, data, foo=4)
    
    # Check set_data using offset in data buffer
    # ------------------------------------------
    def test_set_data_offset(self):
        data = np.zeros(100, np.float32)
        subdata = data[:10]
        
        B = DataBuffer(data)
        B.set_subdata(subdata, offset=10)
        last_cmd = B._glir.clear()[-1]
        offset = last_cmd[2]
        assert offset == 10*4
    
    def test_getitem(self):
        dtype = np.dtype([('position', np.float32, 3),
                          ('texcoord', np.float32, 2),
                          ('color',    np.float32, 4)])
        data = np.zeros(10, dtype=dtype)
        B = DataBuffer(data)
        assert B[1].dtype == dtype
        assert B[1].size == 1
        assert B[-1].dtype == dtype
        assert B[-1].size == 1
        
        self.assertRaises(IndexError, B.__getitem__, +999)
        self.assertRaises(IndexError, B.__getitem__, -999)
    
    def test_setitem(self):
        dtype = np.dtype([('position', np.float32, 3),
                          ('texcoord', np.float32, 2),
                          ('color',    np.float32, 4)])
        data = np.zeros(10, dtype=dtype)
        B = DataBuffer(data)
        B[1] = data[0]
        B[-1] = data[0]
        B[:5] = data[:5]
        B[5:0] = data[:5]  # Weird, but we apparently support this
        # the below is no longer supported starting with NumPy 1.14.0
        # this used to be converted to a single element array of the above
        # dtype filled with 0s. This is no longer supported.
        # B[1] = b''  # Gets converted into array of dtype. Lists do not work
        # the below doesn't work on all systems (I guess?)
        # B[1] = 0
        B[1] = ([0, 0, 0], [0, 0], [0, 0, 0, 0],)
        
        self.assertRaises(IndexError, B.__setitem__, +999, data[0])
        self.assertRaises(IndexError, B.__setitem__, -999, data[0])
        self.assertRaises(TypeError, B.__setitem__, [], data[0])
        
    # Setitem + broadcast
    # ------------------------------------------------------
    def test_setitem_broadcast(self):
        dtype = np.dtype([('position', np.float32, 3),
                          ('texcoord', np.float32, 2),
                          ('color',    np.float32, 4)])
        data = np.zeros(10, dtype=dtype)
        B = DataBuffer(data)
        self.assertRaises(ValueError, B.__setitem__, 'position', (1, 2, 3))

    # Set every 2 item
    # ------------------------------------------------------
    def test_setitem_strided(self):
        dtype = np.dtype([('position', np.float32, 3),
                          ('texcoord', np.float32, 2),
                          ('color',    np.float32, 4)])
        data1 = np.zeros(10, dtype=dtype)
        data2 = np.ones(10, dtype=dtype)
        B = DataBuffer(data1)
        s = slice(None, None, 2)
        self.assertRaises(ValueError, B.__setitem__, s, data2[::2])

    # Set half the array
    # ------------------------------------------------------
    def test_setitem_half(self):
        dtype = np.dtype([('position', np.float32, 3),
                          ('texcoord', np.float32, 2),
                          ('color',    np.float32, 4)])
        data1 = np.zeros(10, dtype=dtype)
        data2 = np.ones(10, dtype=dtype)
        B = DataBuffer(data1)
        B._glir.clear()
        B[:5] = data2[:5]
        glir_cmds = B._glir.clear()
        assert len(glir_cmds) == 1
        set_data = glir_cmds[0][-1]
        assert np.allclose(set_data['position'], data2['position'][:5])
        assert np.allclose(set_data['texcoord'][:5], data2['texcoord'][:5])
        assert np.allclose(set_data['color'][:5], data2['color'][:5])

    # Set field without storage: error
    # --------------------------------
    def test_setitem_field_no_storage(self):
        dtype = np.dtype([('position', np.float32, 3),
                          ('texcoord', np.float32, 2),
                          ('color',    np.float32, 4)])
        data = np.zeros(10, dtype=dtype)
        B = DataBuffer(data)
        self.assertRaises(ValueError,  B.__setitem__, 'position', (1, 2, 3))

    # Set every 2 item without storage:  error
    # ----------------------------------------
    def test_every_two_item_no_storage(self):
        dtype = np.dtype([('position', np.float32, 3),
                          ('texcoord', np.float32, 2),
                          ('color',    np.float32, 4)])
        data = np.zeros(10, dtype=dtype)
        B = DataBuffer(data)
        # with self.assertRaises(ValueError):
        #    B[::2] = data[::2]
        s = slice(None, None, 2)
        self.assertRaises(ValueError, B.__setitem__, s, data[::2])

    # Resize
    # ------
    def test_resize(self):
        data = np.zeros(10)
        B = DataBuffer(data=data)
        data = np.zeros(20)
        B.set_data(data)
        assert B.nbytes == data.nbytes

    # Resize now allowed using ellipsis
    # -----------------------------
    def test_no_resize_ellipsis(self):
        data = np.zeros(10)
        B = DataBuffer(data=data)
        data = np.zeros(30)
        self.assertRaises(ValueError, B.__setitem__, Ellipsis, data)
        
    # Broadcast when using ellipses
    def test_broadcast_ellipsis(self):
        data = np.zeros(10)
        B = DataBuffer(data=data)
        data = np.zeros(5)
        B[Ellipsis] = data
        glir_cmd = B._glir.clear()[-1]
        assert glir_cmd[-1].shape == (10,)


class DataBufferViewTest(unittest.TestCase):
    
    def test_init_view(self):
        data = np.zeros(10)
        B = DataBuffer(data=data)
        
        V = DataBufferView(B, 1)
        assert V.size == 1
        
        V = DataBufferView(B, slice(0, 5))
        assert V.size == 5
        
        V = DataBufferView(B, slice(5, 0))
        assert V.size == 5
        
        V = DataBufferView(B, Ellipsis)
        assert V.size == 10
        
        self.assertRaises(TypeError, DataBufferView, B, [])
        self.assertRaises(ValueError, DataBufferView, B, slice(0, 10, 2))
        
        
# -----------------------------------------------------------------------------
class VertexBufferTest(unittest.TestCase):

    # VertexBuffer allowed base types
    # -------------------------------
    def test_init_allowed_dtype(self):
        for dtype in (np.uint8, np.int8, np.uint16, np.int16, np.float32):
            V = VertexBuffer(np.zeros((10, 3), dtype=dtype))
            names = V.dtype.names
            assert V.dtype[names[0]].base == dtype
            assert V.dtype[names[0]].shape == (3,)
        for dtype in (np.float64, np.int64):
            self.assertRaises(TypeError, VertexBuffer,
                              np.zeros((10, 3), dtype=dtype))

        # Tuple/list is also allowed
        V = VertexBuffer([1, 2, 3])
        assert V.size == 3
        assert V.itemsize == 4
        #
        V = VertexBuffer([[1, 2], [3, 4], [5, 6]])
        assert V.size == 3
        assert V.itemsize == 2 * 4
        
        # Convert
        data = np.zeros((10,), 'uint8')
        B = VertexBuffer(data)
        assert B.dtype[0].base == np.uint8
        assert B.dtype[0].itemsize == 1
        #
        data = np.zeros((10, 2), 'uint8')
        B = VertexBuffer(data)
        assert B.dtype[0].base == np.uint8
        assert B.dtype[0].itemsize == 2
        B.set_data(data, convert=True)
        assert B.dtype[0].base == np.float32
        assert B.dtype[0].itemsize == 8
        B = VertexBuffer(data[::2].copy())
        
        # This is converted to 1D
        B = VertexBuffer([[1, 2, 3, 4, 5], [1, 2, 3, 4, 5]])
        assert B.size == 10 
         
        # Not allowed
        self.assertRaises(TypeError, VertexBuffer, dtype=np.float64)
        #self.assertRaises(TypeError, VertexBuffer, [[1,2,3,4,5],[1,2,3,4,5]])

    # VertexBuffer not allowed base types
    # -----------------------------------
    def test_init_not_allowed_dtype(self):
        for dtype in (np.uint32, np.int32, np.float64):
            # with self.assertRaises(TypeError):
            #    V = VertexBuffer(dtype=dtype)
            self.assertRaises(TypeError, VertexBuffer, dtype=dtype)

    def test_glsl_type(self):
        
        data = np.zeros((10,), np.float32)
        B = VertexBuffer(data)
        C = B[1:]
        assert B.glsl_type == ('attribute', 'float')
        assert C.glsl_type == ('attribute', 'float')
        
        data = np.zeros((10, 2), np.float32)
        B = VertexBuffer(data)
        C = B[1:]
        assert B.glsl_type == ('attribute', 'vec2')
        assert C.glsl_type == ('attribute', 'vec2')
        
        data = np.zeros((10, 4), np.float32)
        B = VertexBuffer(data)
        C = B[1:]
        assert B.glsl_type == ('attribute', 'vec4')
        assert C.glsl_type == ('attribute', 'vec4')


# -----------------------------------------------------------------------------
class IndexBufferTest(unittest.TestCase):

    # IndexBuffer allowed base types
    # ------------------------------
    def test_init_allowed_dtype(self):
        
        # allowed dtypes
        for dtype in (np.uint8, np.uint16, np.uint32):
            b = IndexBuffer(np.zeros(10, dtype=dtype))
            b.dtype == dtype

        # no data => no dtype
        V = IndexBuffer()
        V.dtype is None
        
        # Not allowed dtypes
        for dtype in (np.int8, np.int16, np.int32,
                      np.float16, np.float32, np.float64):
            # with self.assertRaises(TypeError):
            #    V = IndexBuffer(dtype=dtype)
            data = np.zeros(10, dtype=dtype)
            self.assertRaises(TypeError, IndexBuffer, data)
        
        # Prepare some data
        dtype = np.dtype([('position', np.float32, 3),
                          ('texcoord', np.float32, 2), ])
        sdata = np.zeros(10, dtype=dtype)
        
        # Normal data is
        data = np.zeros([1, 2, 3], np.uint8)
        B = IndexBuffer(data)
        assert B.dtype == np.uint8
        
        # We can also convert
        B.set_data(data, convert=True)
        assert B.dtype == np.uint32
        
        # Structured data not allowed
        self.assertRaises(TypeError, IndexBuffer, dtype=dtype)
        self.assertRaises(TypeError, B.set_data, sdata)


run_tests_if_main()



# -*- coding: utf-8 -*-
# Copyright (c) Vispy Development Team. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.

from ..shaders import Function
from ..transforms import NullTransform
from ...geometry import Rect


clip_frag = """
void clip() {
    vec4 pos = $fb_to_clip(gl_FragCoord);
    if( pos.x < $view.x || pos.x > $view.y || 
        pos.y < $view.z || pos.y > $view.w ) {
        discard;
    }
}
"""


class Clipper(object):
    """Clips visual output to a rectangular region.
    """
    def __init__(self, bounds=(0, 0, 1, 1), transform=None):
        self.clip_shader = Function(clip_frag)
        self.clip_expr = self.clip_shader()
        self.bounds = bounds  # (x, y, w, h)
        if transform is None:
            transform = NullTransform()
        self._transform = None
        self.transform = transform
    
    @property
    def bounds(self):
        """The clipping boundaries.
        
        This must be a tuple (x, y, w, h) in a clipping coordinate system
        that is defined by the `transform` property.
        """
        return self._bounds
    
    @bounds.setter
    def bounds(self, b):
        self._bounds = Rect(b).normalized()
        b = self._bounds
        self.clip_shader['view'] = (b.left, b.right, b.bottom, b.top)

    @property
    def transform(self):
        """The transform that maps from framebuffer coordinates to clipping
        coordinates.
        """
        return self._transform

    @transform.setter
    def transform(self, tr):
        if tr is self._transform:
            return
        self._transform = tr
        self.clip_shader['fb_to_clip'] = tr
        
    def _attach(self, visual):
        try:
            hook = visual._get_hook('frag', 'pre')
        except KeyError:
            raise NotImplementedError("Visual %s does not support clipping" %
                                      visual)
        hook.add(self.clip_expr, position=1)

    def _detach(self, visual):
        visual._get_hook('frag', 'pre').remove(self.clip_expr)



import sys
from unittest import SkipTest

import numpy as np
from numpy.testing import assert_array_almost_equal

from vispy.testing import run_tests_if_main
from vispy.geometry.triangulation import Triangulation as T


def assert_array_eq(a, b):
    assert a.shape == b.shape
    assert a.dtype == b.dtype
    mask = np.isnan(a)
    assert np.all(np.isnan(b[mask]))
    assert np.all(a[~mask] == b[~mask])


def test_intersect_edge_arrays():
    global t
    pts = np.array([
        [0., 0.],
        [0., 10.],
        [5., 0.],
        [-5., 0.],
        [-1., 11.],
        [1., 9.],
    ])
    edges = np.array([
        [0, 1],
        [2, 3],
        [0, 3],
        [4, 5],
        [4, 1],
        [0, 1],
    ])

    lines = pts[edges]
    t = T(pts, edges)

    # intersect array of one edge with a array of many edges
    intercepts = t._intersect_edge_arrays(lines[0:1], lines[1:])
    expect = np.array([0.5, 0.0, 0.5, 1.0, np.nan])
    assert_array_eq(intercepts, expect)

    # intersect every line with every line
    intercepts = t._intersect_edge_arrays(lines[:, np.newaxis, ...],
                                          lines[np.newaxis, ...])
    for i in range(lines.shape[0]):
        int2 = t._intersect_edge_arrays(lines[i], lines)
        assert_array_eq(intercepts[i], int2)


def test_edge_intersections():
    global t
    pts = np.array([
        [0, 0],
        [1, 0],
        [1, 1],
        [0, 1],
        [0, 0.5],  # three edges intersect here
        [2, 0.5],
        [-1, 0.2],
        [2, 0.8],
        [-1, 1],
        [0, 0.5],
    ])
    edges = np.array([
        [0, 1],
        [1, 2],
        [2, 3],
        [3, 0],
        [4, 5],
        [6, 7],
        [8, 9],
    ])

    t = T(pts, edges)

    # first test find_edge_intersections
    cuts = t._find_edge_intersections()
    expect = {
        0: [],
        1: [(0.5, [1., 0.5]),
            (0.6, [1., 0.6])],
        2: [],
        3: [(0.5, [0., 0.5]),
            (0.6, [0., 0.4])],
        4: [(0.25, [0.5, 0.5]),
            (0.5, [1., 0.5])],
        5: [(1./3., [0., 0.4]),
            (0.5, [0.5, 0.5]),
            (2./3., [1., 0.6])],
    }

    assert len(expect) == len(cuts)
    for k, v in expect.items():
        assert len(v) == len(cuts[k])
        for i, ecut in enumerate(v):
            vcut = cuts[k][i]
            assert len(vcut) == len(ecut)
            for j in range(len(vcut)):
                assert_array_almost_equal(np.array(ecut[j]), np.array(vcut[j]))

    # next test that we can split the edges correctly
    t._split_intersecting_edges()
    pts = np.array([[0., 0.],
                    [1., 0.],
                    [1., 1.],
                    [0., 1.],
                    [0., 0.5],
                    [2., 0.5],
                    [-1., 0.2],
                    [2., 0.8],
                    [-1., 1.],
                    [0., 0.5],
                    [1., 0.5],
                    [1., 0.6],
                    [0., 0.5],
                    [0., 0.4],
                    [0.5, 0.5],
                    [1., 0.5],
                    [0., 0.4],
                    [0.5, 0.5],
                    [1., 0.6]])
    edges = np.array([[0, 1],
                      [1, 10],
                      [2, 3],
                      [3, 12],
                      [4, 14],
                      [6, 16],
                      [8, 9],
                      [10, 11],
                      [11, 2],
                      [12, 13],
                      [13, 0],
                      [14, 15],
                      [15, 5],
                      [16, 17],
                      [17, 18],
                      [18, 7]])

    if sys.version[0] == '3':
        raise SkipTest('Triangulation differences on Py3k')
    assert_array_almost_equal(pts, t.pts)
    assert np.all(edges == t.edges)

    # Test _nearly_ parallel lines.
    pts = np.array([[0., 0.],
                    [1.62434542,  0.],
                    [1.62434542, -0.61175638],
                    [1.09617364, -0.61175638]])

    edges = np.array([[0, 1], [1, 2], [2, 3], [3, 0]])
    t = T(pts, edges)
    for edge, cuts in t._find_edge_intersections().items():
        assert len(cuts) == 0


def test_merge_duplicate_points():
    global t
    pts = np.array([
        [0, 0],
        [1, 1],
        [0.1, 0.7],
        [2, 3],
        [0, 0],
        [0.1, 0.7],
        [5, 6],
    ])
    edges = np.array([
        [0, 6],
        [1, 5],
        [2, 4],
        [3, 6],
        [4, 5],
    ])

    t = T(pts, edges)
    t._merge_duplicate_points()

    pts = np.array([
        [0, 0],
        [1, 1],
        [0.1, 0.7],
        [2, 3],
        [5, 6],
    ])
    edges = np.array([
        [0, 4],
        [1, 2],
        [2, 0],
        [3, 4],
        [0, 2],
    ])
    assert np.allclose(t.pts, pts)
    assert np.all(t.edges == edges)


def test_initialize():
    # check points are correctly sorted
    # check artificial points are outside bounds of all others
    # check tops / bottoms
    pass


def test_utility_methods():
    global t
    pts = np.array([
        [0, 0],
        [1, 0],
        [2, 0],
        [3, 0],
        [1.5, 2],
        [1.5, -2],
    ])
    edges = np.array([
        [4, 5],  # edge cuts through triangle (1, 2, 4)
    ])

    t = T(pts, edges)
    # skip initialization and just simulate being part-way through
    # triangulation
    for tri in [[0, 1, 4], [1, 2, 4], [2, 3, 4]]:
        t._add_tri(*tri)

    # find_cut_triangle
    assert t._find_cut_triangle((4, 5)) == (4, 1, 2)

    # orientation
    assert t._orientation((4, 5), 0) == 1
    assert t._orientation((4, 5), 1) == 1
    assert t._orientation((4, 5), 2) == -1
    assert t._orientation((4, 5), 3) == -1
    assert t._orientation((4, 5), 4) == 0
    assert t._orientation((4, 5), 5) == 0

    # distance
    dist = ((t.pts[0]-t.pts[1])**2).sum()**0.5
    assert t._distance(t.pts[0], t.pts[1]) == dist

    # adjacent_tri
    assert t._adjacent_tri((1, 4), 0) == (4, 1, 2)
    assert t._adjacent_tri((0, 4), 1) is None
    assert t._adjacent_tri((1, 4), (1, 4, 0)) == (4, 1, 2)
    assert t._adjacent_tri((0, 4), (1, 4, 0)) is None
    try:
        t._adjacent_tri((1, 4), 5)
    except RuntimeError:
        pass
    else:
        raise Exception("Expected RuntimeError.")

    # edges_intersect
    assert not t._edges_intersect((0, 1), (1, 2))
    assert not t._edges_intersect((0, 2), (1, 2))
    assert t._edges_intersect((4, 5), (1, 2))

    # is_constraining_edge
    assert t._is_constraining_edge((4, 5))
    assert t._is_constraining_edge((5, 4))
    assert not t._is_constraining_edge((3, 5))
    assert not t._is_constraining_edge((3, 2))


def test_projection():
    pts = np.array([[0, 0],
                    [5, 0],
                    [1, 2],
                    [3, 4]])
    t = T(pts, np.zeros((0, 2)))

    a, b, c, d = pts
    assert np.allclose(t._projection(a, c, b), [1, 0])
    assert np.allclose(t._projection(b, c, a), [1, 0])
    assert np.allclose(t._projection(a, d, b), [3, 0])
    assert np.allclose(t._projection(b, d, a), [3, 0])
    assert np.allclose(t._projection(a, b, c), [1, 2])
    assert np.allclose(t._projection(c, b, a), [1, 2])


def test_random():
    # Just test that these triangulate without exception.
    # TODO: later on, we can turn this same test into an image comparison
    # with Polygon.
    N = 10
    np.random.seed(0)

    for i in range(4):
        pts = np.random.normal(size=(N, 2))
        edges = np.zeros((N, 2), dtype=int)
        edges[:, 0] = np.arange(N)
        edges[:, 1] = np.arange(1, N+1) % N

        t = T(pts, edges)
        t.triangulate()

    theta = np.linspace(0, 2*np.pi, 11)[:-1]
    pts = np.hstack([np.cos(theta)[:, np.newaxis],
                    np.sin(theta)[:, np.newaxis]])
    pts[::2] *= 0.4
    edges = np.empty((pts.shape[0], 2), dtype=np.uint)
    edges[:, 0] = np.arange(pts.shape[0])
    edges[:, 1] = edges[:, 0] + 1
    edges[-1, 1] = 0
    t = T(pts, edges)
    t.triangulate()

    # much larger test
    # this should pass, but takes forever..
    #N = 4000
    #pts = np.random.normal(size=(N, 2))
    #pts = np.cumsum(pts, axis=0)
    #edges = np.zeros((N, 2), dtype=int)
    #edges[:,0] = np.arange(N)
    #edges[:,1] = np.arange(1,N+1) % N

    #t = T(pts, edges)
    #t.triangulate()


def test_orthogonal():
    # make lines that are entirely vertical / horizontal
    np.random.seed(1)
    N = 100
    pts = [[0, 0]]
    for i in range(N - 1):
        p = pts[-1][:]
        p[i % 2] += np.random.normal()
        pts.append(p)
    pts = np.array(pts)
    edges = np.zeros((N, 2), dtype=int)
    edges[:, 0] = np.arange(N)
    edges[:, 1] = np.arange(1, N + 1) % N

    t = T(pts, edges)
    t.triangulate()


def test_edge_event():
    # mode 1
    pts = np.array([[0, 0],
                    [5, -10],
                    [10, 0],
                    [6, -5],
                    [5, 5],
                    ])
    inds = np.arange(pts.shape[0])[:, np.newaxis]
    edges = np.hstack([inds, np.roll(inds, -1)])

    t = T(pts, edges)
    t.triangulate()

    t = T(pts * [-1, 1], edges)
    t.triangulate()

    # mode 2
    pts = np.array([[0, 0],
                    [10, 0],
                    [20, 0],
                    [5, 11],
                    ])
    inds = np.arange(pts.shape[0])[:, np.newaxis]
    edges = np.hstack([inds, np.roll(inds, -1)])

    t = T(pts, edges)
    t.triangulate()

    t = T(pts * [-1, 1], edges)
    t.triangulate()

    # mode 1, 2
    pts = np.array([[0, 0],
                    [10, 0],
                    [20, 0],
                    [5, 11],
                    [9, 10],
                    [0, 20],
                    ])
    inds = np.arange(pts.shape[0])[:, np.newaxis]
    edges = np.hstack([inds, np.roll(inds, -1)])

    t = T(pts, edges)
    t.triangulate()

    t = T(pts * [-1, 1], edges)
    t.triangulate()

    # mode 2, 1
    pts = np.array([[0, 0],
                    [10, 0],
                    [20, 0],
                    [15, 8],
                    [15, 1],
                    [-5, 10],
                    ])
    inds = np.arange(pts.shape[0])[:, np.newaxis]
    edges = np.hstack([inds, np.roll(inds, -1)])

    t = T(pts, edges)
    t.triangulate()

    t = T(pts * [-1, 1], edges)
    t.triangulate()

    # mode 2, 1 with many triangles
    pts = np.array([[0, 10],
                    [2, 8],
                    [4, 6],
                    [6, 4],
                    [8, 2],
                    [10, 0],

                    [20, 5],
                    [20, 20],

                    [2, 13],
                    [4, 11],
                    [6, 9],
                    [8, 7],
                    [10, 5],

                    [10, 1],
                    [0, 15],
                    ])
    inds = np.arange(pts.shape[0])[:, np.newaxis]
    edges = np.hstack([inds, np.roll(inds, -1)])

    t = T(pts, edges)
    t.triangulate()

    t = T(pts * [-1, 1], edges)
    t.triangulate()

    # mode 1, 2, 1, 2, 1
    pts = np.array([[0, 10],
                    [2, 9],
                    [4, 8],
                    [6, 7],
                    [8, 6],
                    [10, 5],

                    [20, 5],
                    [20, 20],

                    [2, 11],
                    [19, 19],
                    [6, 9],
                    [19, 18],
                    [10, 7],

                    [11, 5.1],
                    [0, 11.1],
                    ])
    inds = np.arange(pts.shape[0])[:, np.newaxis]
    edges = np.hstack([inds, np.roll(inds, -1)])

    t = T(pts, edges)
    t.triangulate()

    t = T(pts * [-1, 1], edges)
    t.triangulate()

    # mode 2, 1, 2, 1
    pts = np.array([[0, 10],
                    [2, 9],
                    [4, 8],
                    [6, 7],
                    [8, 6],
                    [10, 5],

                    [20, 5],
                    [20, 20],

                    [6, 9],
                    [19, 18],
                    [10, 7],

                    [11, 5.1],
                    [0, 11.1],
                    ])
    inds = np.arange(pts.shape[0])[:, np.newaxis]
    edges = np.hstack([inds, np.roll(inds, -1)])

    t = T(pts, edges)
    t.triangulate()

    t = T(pts * [-1, 1], edges)
    t.triangulate()

    # 1, 2  upper/lower polygon order check
    pts = np.array([[-5, 0],
                    [-3, 0],
                    [10, 0],
                    [15, 15],
                    [4, 9],
                    [6, 8.8],
                    [9, 10],
                    ])
    inds = np.arange(pts.shape[0])[:, np.newaxis]
    edges = np.hstack([inds, np.roll(inds, -1)])

    t = T(pts, edges)
    t.triangulate()

    t = T(pts * [-1, 1], edges)
    t.triangulate()


run_tests_if_main()



from vispy.testing import run_tests_if_main
from vispy.app import set_interactive
from vispy.ext.ipy_inputhook import inputhook_manager


# Expect the inputhook_manager to set boolean `_in_event_loop`
# on instances of this class when enabled.
class MockApp(object):
    pass


def test_interactive():
    f = MockApp()
    set_interactive(enabled=True, app=f)

    assert inputhook_manager._current_gui == 'vispy'
    assert f._in_event_loop
    assert 'vispy' in inputhook_manager.apps
    assert f == inputhook_manager.apps['vispy']

    set_interactive(enabled=False)

    assert inputhook_manager._current_gui is None
    assert not f._in_event_loop


run_tests_if_main()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) Vispy Development Team. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
# -----------------------------------------------------------------------------
from vispy import scene, io
from vispy.testing import (requires_application, TestingCanvas,
                           run_tests_if_main)
from vispy.testing.image_tester import assert_image_approved


@requires_application()
def test_perspective_render():
    with TestingCanvas(size=(120, 200)) as canvas:

        grid = canvas.central_widget.add_grid()
        grid.padding = 20

        imdata = io.load_crate().astype('float32') / 255

        views = []
        images = []
        for i, imethod in enumerate(['impostor', 'subdivide']):
            v = grid.add_view(row=i, col=0, border_color='white')
            v.camera = 'turntable'
            v.camera.fov = 50
            v.camera.distance = 30

            views.append(v)
            image = scene.visuals.Image(imdata, method=imethod,
                                        grid=(4, 4))
            image.transform = scene.STTransform(translate=(-12.8, -12.8),
                                                scale=(0.1, 0.1))
            v.add(image)
            images.append(image)

        image = canvas.render()
        canvas.close()

        # Allow many pixels to differ by a small amount--texture sampling and
        # exact triangle position will differ across platforms. However a
        # change in perspective or in the widget borders should trigger a
        # failure.
        assert_image_approved(image, 'scene/cameras/perspective_test.png',
                              'perspective test 1: 2 identical views with '
                              'correct perspective',
                              px_threshold=20,
                              px_count=60,
                              max_px_diff=200)


run_tests_if_main()
